import 'dart:async';
import 'dart:io';

import 'package:connectivity_plus/connectivity_plus.dart';

import '../services/event_bus.dart';
import 'utils.dart';

/// Listen network status and broadcast [ConnectivityEvent] event via [eventBus].
///
/// ```
/// // Start listener
/// ConnectivityListener.startup(['google.com', 'youtube.com']);
///
/// // Listen ConnectivityState
/// var connectivityEventStream = eventBus.on<ConnectivityEvent>().listen((event) {
///   if (event.state == ConnectivityState.WIFI_AVAILABLE) {
///     ...
///   }
/// });
/// connectivityEventStream.cancel();
/// ```
class ConnectivityListener {
  static ConnectivityListener? _instance;

  late _Availability _availability;
  late StreamSubscription _listener;
  List<String>? _availabilityHosts;

  /// Start network status listen (single)
  ///
  /// If [availabilityHosts] is List<String>, detect from the first one until
  /// it succeeds.
  ///
  /// If [availabilityHosts] is null, only [ConnectivityState.WIFI],
  /// [ConnectivityState.MOBILE], [ConnectivityState.NONE] three events will be
  /// broadcast.
  ///
  factory ConnectivityListener.startup({
    List<String>? availabilityHosts,
  }) {
    return _instance ??= ConnectivityListener._(availabilityHosts);
  }

  ConnectivityListener._(this._availabilityHosts) {
    /// listener
    _availability = _Availability();
    _listener = Connectivity()
        .onConnectivityChanged
        .listen((ConnectivityResult result) async {
      var res = enumFromString(
          ConnectivityState.values, enumToString(result).toUpperCase());
      eventBus.fire(ConnectivityEvent(res));

      if (availabilityHosts != null && res != ConnectivityState.NONE) {
        var availabilityRes =
            await _availability.check(res, availabilityHosts!);
        if (availabilityRes != null) {
          eventBus.fire(ConnectivityEvent(res));
        }
      }
    });
  }

  /// Stop listener
  void stop() {
    _listener.cancel();
    _instance = null;
  }

  /// Set the [_availabilityHosts] of the single instance
  static set availabilityHosts(List<String>? availabilityHosts) {
    if (_instance != null) {
      _instance!._availabilityHosts = availabilityHosts;
    }
  }

  /// Get the [_availabilityHosts] of the single instance
  static List<String>? get availabilityHosts {
    if (_instance != null) {
      return _instance!._availabilityHosts;
    }
    return null;
  }

  /// Sample usage to check current state.
  ///
  /// If [availabilityHosts] is List<String>, detect from the first one until
  /// it succeeds.
  ///
  /// If [availabilityHosts] is null, [ConnectivityListener.availabilityHosts]
  /// will be used. if both are null, only return [ConnectivityState.WIFI],
  /// [ConnectivityState.MOBILE], [ConnectivityState.NONE] three states.
  ///
  /// ```
  /// final state = await ConnectivityListener.checkState();
  /// // ------ or ------
  /// ConnectivityListener.startup(
  ///   availabilityHosts: ['google.com', 'youtube.com'],
  /// );
  /// final state = await ConnectivityListener.checkState();
  /// // ------ or ------
  /// final state = await ConnectivityListener.checkState(
  ///   availabilityHosts:['google.com', 'youtube.com'],
  /// );
  /// ```
  static Future<ConnectivityState> checkState(
      {List<String>? availabilityHosts}) async {
    final res = enumFromString(ConnectivityState.values,
        enumToString(await (Connectivity().checkConnectivity())).toUpperCase());

    if (availabilityHosts == null) {
      availabilityHosts = ConnectivityListener.availabilityHosts;
    }

    if (availabilityHosts != null && res != ConnectivityState.NONE) {
      return (await _Availability().check(res, availabilityHosts))!;
    }

    return res;
  }
}

class _Availability {
  String? _tokenKey;

  Future<ConnectivityState?> check(
      ConnectivityState preState, List<String> hosts) async {
    var tokenKey = _createTokenKey();
    _tokenKey = tokenKey;

    for (int i = 0; i < hosts.length; i++) {
      final result = await InternetAddress.lookup(hosts[i]);
      if (_tokenKey != tokenKey) {
        return null;
      }
      if (result.isNotEmpty && result[0].rawAddress.isNotEmpty) {
        return enumFromString(
            ConnectivityState.values, '${enumToString(preState)}_AVAILABLE');
      }
    }

    return enumFromString(
        ConnectivityState.values, '${enumToString(preState)}_NOT_AVAILABLE');
  }

  String _createTokenKey() {
    return DateTime.now().millisecondsSinceEpoch.toString();
  }
}

class ConnectivityEvent {
  final ConnectivityState state;

  ConnectivityEvent(this.state);
}

enum ConnectivityState {
  /// Device connected via Wi-Fi，but availability has not been detected.
  WIFI,

  /// Device connected via Wi-Fi，and the availability has been tested.
  WIFI_AVAILABLE,

  /// Device connected via Wi-Fi，and the availability test is not passed.
  WIFI_NOT_AVAILABLE,

  /// Device connected to cellular network，but availability has not been detected.
  MOBILE,

  /// Device connected to cellular network，and the availability has been tested.
  MOBILE_AVAILABLE,

  /// Device connected to cellular network，and the availability test is not passed.
  MOBILE_NOT_AVAILABLE,

  /// evice not connected to any network
  NONE,
}
